@@ -7,7 +7,7 @@ using PanelApplet;
 
 namespace Tomboy
 {
-	public class TomboyApplet : PanelApplet
+	public class TomboyApplet : PanelApplet.PanelApplet
 	{
 		NoteManager manager;
 		TomboyTray tray;
@@ -62,46 +62,53 @@ namespace Tomboy
 					   "GNOME_TomboyApplet.xml",
 					   "Tomboy",
 					   menu_verbs);
-
-			// FIXME: Connecting to this crashes in the C# bindings.
-			//ChangeBackground += OnChangeBackgroundEvent;
 		}
 
 		void ShowPreferencesVerb ()
 		{
-			tray.ShowPreferences ();
+			if (tray != null)
+				tray.ShowPreferences ();
 		}
 
 		void ShowPluginsVerb ()
 		{
-			manager.PluginManager.ShowPluginsDirectory ();
+			if (manager != null)
+				manager.PluginManager.ShowPluginsDirectory ();
 		}
 
 		void ShowAboutVerb ()
 		{
-			tray.ShowAbout ();
+			if (tray != null)
+				tray.ShowAbout ();
 		}
 
-		void OnChangeBackgroundEvent (object sender, ChangeBackgroundArgs args)
+		// NOTE: This is needed to support transparent/pixmap panel
+		//       backgrounds correctly, since Gtk# doesn't map
+		//       GtkStyle::bg_pixmaps[].
+		[DllImport ("libtomboy")]
+		private static extern IntPtr tomboy_widget_set_bg_pixmap (IntPtr widget, IntPtr pixmap);
+
+		// FIXME: Overriding to this crashes in the C# bindings.
+		//        Manually tweaked generated bindings to not crash.
+		protected override void OnChangeBackground (BackgroundType type, 
+							    Gdk.Color      color, 
+							    Gdk.Pixmap     pixmap)
 		{
-			// This is needed to support transparent panel
-			// backgrounds correctly.
-
-			Console.WriteLine ("OnChangeBackgroundEvent Called!");
-
-			switch (args.Type) {
-			case BackgroundType.NoBackground:
-			case BackgroundType.PixmapBackground:
+			if (tray != null) {
 				Gtk.RcStyle rc_style = new Gtk.RcStyle ();
+				tray.Style = null;
+				tray.ModifyStyle (rc_style);
 
-				tray.Image.ModifyStyle (rc_style);
-				ModifyStyle (rc_style);
-				break;
-
-			case BackgroundType.ColorBackground:
-				tray.Image.ModifyBg (Gtk.StateType.Normal, args.Color);
-				ModifyBg (Gtk.StateType.Normal, args.Color);
-				break;
+				switch (type) {
+				case BackgroundType.ColorBackground:
+					tray.ModifyBg (Gtk.StateType.Normal, color);
+					break;
+				case BackgroundType.NoBackground:
+					break;
+				case BackgroundType.PixmapBackground:
+					tomboy_widget_set_bg_pixmap (tray.Handle, pixmap.Handle);
+					break;
+				}
 			}
 		}
 	}
