@@ -1,5 +1,7 @@
 
 using System;
+using System.Runtime.InteropServices;
+using System.Text;
 using Mono.Posix;
 
 namespace Tomboy
@@ -13,7 +15,8 @@ namespace Tomboy
 			LeftMargin = 8;
 			RightMargin = 8;
 			CanDefault = true;
-			DragDataReceived += OnDragDataReceived;
+
+			// Make sure the cursor position is visible
 			ScrollMarkOnscreen (buffer.InsertMark);
 
 			// Set Font from GConf preference
@@ -25,11 +28,21 @@ namespace Tomboy
 
 			// Set extra editor drag targets supported (in addition
 			// to the default TextView's various text formats)...
-			Gtk.TargetList list = Gtk.Drag.DestGetTargetList (this);
+			//// Gtk.TargetList list = Gtk.Drag.DestGetTargetList (this);
+
+			IntPtr list_ptr = gtk_drag_dest_get_target_list (this.Handle);
+			Gtk.TargetList list = new Gtk.TargetList (list_ptr);
+
 			list.Add (Gdk.Atom.Intern ("text/uri-list", false), 0, 1);
 			list.Add (Gdk.Atom.Intern ("_NETSCAPE_URL", false), 0, 1);
 		}
 
+		// FIXME: Gtk# broke compatibility at some point with the
+		// methodref for DestGetTargetList.  We invoke it manually so
+		// our binary will work for everyone.
+		[DllImport("libgtk-win32-2.0-0.dll")]
+		static extern IntPtr gtk_drag_dest_get_target_list (IntPtr raw);
+
 		//
 		// Update the font based on the changed Preference dialog setting.
 		//
@@ -60,46 +73,67 @@ namespace Tomboy
 		//
 		// DND Drop handling
 		//
-		void OnDragDataReceived (object sender, Gtk.DragDataReceivedArgs args)
+		protected override void OnDragDataReceived (Gdk.DragContext context, 
+							    int x,
+							    int y,
+							    Gtk.SelectionData selection_data,
+							    uint info,
+							    uint time)
 		{
-			// FIXME: Need to access SelectionData's target, which
-			// is unmapped in Gtk# currently, and only insert urls
-			// when the target is text/uri-list, or _NETSCAPE_URL.
+			bool has_url = false;
 
-			UriList uri_list = new UriList (args.SelectionData);
-			if (uri_list.Count == 0)
-				return;
+			foreach (Gdk.Atom target in context.Targets) {
+				if (target.Name == "text/uri-list" ||
+				    target.Name == "_NETSCAPE_URL") {
+					has_url = true;
+					break;
+				}
+			}
 
-			Gtk.TextIter cursor = Buffer.GetIterAtMark (Buffer.InsertMark);
-			bool more_than_one = false;
+			if (has_url) {
+				UriList uri_list = new UriList (selection_data);
+				StringBuilder insert = new StringBuilder ();
 
-			foreach (Uri uri in uri_list) {
-				Console.WriteLine ("Got Dropped URI: {0}", uri);
-				if (more_than_one)
-					Buffer.Insert (cursor, "\n");
+				foreach (Uri uri in uri_list) {
+					Console.WriteLine ("Got Dropped URI: {0}", uri);
 
-				if (uri.IsFile) 
-					Buffer.Insert (cursor, uri.LocalPath);
-				else
-					Buffer.Insert (cursor, uri.ToString ());
+					// FIXME: The space here is a hack
+					// around a bug in the URL Regex which
+					// matches across newlines.
+					if (insert.Length > 0)
+						insert.Append (" \n");
 
-				more_than_one = true;
+					if (uri.IsFile) 
+						insert.Append (uri.LocalPath);
+					else
+						insert.Append (uri.ToString ());
+				}
+
+				if (insert.Length > 0)
+					Buffer.Insert (Buffer.GetIterAtMark (Buffer.InsertMark),
+						       insert.ToString ());
+
+				Gtk.Drag.Finish (context, insert.Length > 0, false, time);
+			} else {
+				base.OnDragDataReceived (context, x, y, selection_data, info, time);
 			}
 		}
 	}
 
-	public class NoteWindow : Gtk.Window 
+	public class NoteWindow : ForcedPresentWindow 
 	{
 		Note note;
 
 		Gtk.AccelGroup accel_group;
 		Gtk.Toolbar toolbar;
+		Gtk.Widget link_button;
 		Gtk.Button text_button;
 		NoteTextMenu text_menu;
 		Gtk.TextView editor;
 		Gtk.ScrolledWindow editor_window;
 
 		GlobalKeybinder global_keys;
+		InterruptableTimeout mark_set_timeout;
 
 		static Gdk.Pixbuf stock_notes;
 
@@ -137,6 +171,11 @@ namespace Tomboy
 			editor.PopulatePopup += OnPopulatePopup;
 			editor.Show ();
 
+			// Sensitize the Link toolbar button on text selection
+			mark_set_timeout = new InterruptableTimeout();
+			mark_set_timeout.Timeout += UpdateLinkButtonSensitivity;
+			note.Buffer.MarkSet += OnSelectionMarkSet;
+
 			// FIXME: I think it would be really nice to let the
 			//        window get bigger up till it grows more than
 			//        60% of the screen, and then show scrollbars. 
@@ -274,6 +313,21 @@ namespace Tomboy
 			get { return toolbar; }
 		}
 
+		// 
+		// Sensitize the Link toolbar button on text selection
+		//
+
+		void OnSelectionMarkSet (object sender, Gtk.MarkSetArgs args)
+		{
+			// FIXME: Process in a timeout due to GTK+ bug #172050.
+			mark_set_timeout.Reset (0);
+		}
+
+		void UpdateLinkButtonSensitivity (object sender, EventArgs args)
+		{
+			link_button.Sensitive = (note.Buffer.Selection != null);
+		}
+
 		//
 		// Right-click menu
 		//
@@ -365,38 +419,36 @@ namespace Tomboy
 		Gtk.Toolbar MakeToolbar ()
 		{
 			Gtk.Toolbar toolbar = new Gtk.Toolbar ();
-			toolbar.IconSize = Gtk.IconSize.LargeToolbar;
-			toolbar.ToolbarStyle = Gtk.ToolbarStyle.BothHoriz;
 			toolbar.Tooltips = true;
 
-			Gtk.Widget link = 
-				toolbar.AppendItem (Catalog.GetString ("Link"), 
-						    Catalog.GetString ("Link selected text to a " +
-								       "new note"), 
-						    null, 
-						    new Gtk.Image (Gtk.Stock.JumpTo, 
-								   Gtk.IconSize.LargeToolbar),
-						    new Gtk.SignalFunc (LinkButtonClicked));
-			link.AddAccelerator ("activate",
-					     accel_group,
-					     (uint) Gdk.Key.l, 
-					     Gdk.ModifierType.ControlMask,
-					     Gtk.AccelFlags.Visible);
-
 			Gtk.Widget find = 
-				toolbar.AppendItem (Catalog.GetString ("Search"), 
-						    Catalog.GetString ("Search your notes"),
-						    null, 
-						    new Gtk.Image (Gtk.Stock.Find, 
-								   Gtk.IconSize.LargeToolbar),
-						    new Gtk.SignalFunc (FindButtonClicked));
+				toolbar.AppendItem (
+					Catalog.GetString ("Search"), 
+					Catalog.GetString ("Search your notes"),
+					null, 
+					new Gtk.Image (Gtk.Stock.Find, toolbar.IconSize),
+					new Gtk.SignalFunc (FindButtonClicked));
 			find.AddAccelerator ("activate",
 					     accel_group,
 					     (uint) Gdk.Key.f, 
 					     Gdk.ModifierType.ControlMask,
 					     Gtk.AccelFlags.Visible);
 
-			text_button = new TextToolButton ();
+			link_button = 
+				toolbar.AppendItem (
+					Catalog.GetString ("Link"), 
+					Catalog.GetString ("Link selected text to a new note"), 
+					null, 
+					new Gtk.Image (Gtk.Stock.JumpTo, toolbar.IconSize),
+					new Gtk.SignalFunc (LinkButtonClicked));
+			link_button.Sensitive = (note.Buffer.Selection != null);
+			link_button.AddAccelerator ("activate",
+						    accel_group,
+						    (uint) Gdk.Key.l, 
+						    Gdk.ModifierType.ControlMask,
+						    Gtk.AccelFlags.Visible);
+
+			text_button = new TextToolButton (toolbar);
 			text_button.ButtonPressEvent += TextButtonPress;
 			text_button.Clicked += TextButtonClicked;
 			text_button.Show ();
@@ -412,15 +464,15 @@ namespace Tomboy
 
 			toolbar.AppendSpace ();
 
-			Gtk.Widget delete = 
-				toolbar.AppendItem (Catalog.GetString ("Delete"), 
-						    Catalog.GetString ("Delete this note"), 
-						    null, 
-						    new Gtk.Image (Gtk.Stock.Delete, 
-								   Gtk.IconSize.LargeToolbar),
-						    new Gtk.SignalFunc (DeleteButtonClicked));
+		        Gtk.Widget delete = 
+				toolbar.AppendItem (
+					Catalog.GetString ("Delete"), 
+					Catalog.GetString ("Delete this note"), 
+					null, 
+					new Gtk.Image (Gtk.Stock.Delete, toolbar.IconSize),
+					new Gtk.SignalFunc (DeleteButtonClicked));
 
-			// Don't allow deleting the "Start" or "Recent Changes" notes...
+			// Don't allow deleting the "Start Here" note...
 			if (note.IsSpecial)
 				delete.Sensitive = false;
 
@@ -429,27 +481,36 @@ namespace Tomboy
 
 		class TextToolButton : Gtk.Button
 		{
-			public TextToolButton ()
+			Gtk.Image image;
+			Gtk.Label label_horiz;
+			Gtk.Label label_vert;
+			Gtk.Arrow arrow;
+			Gtk.VBox box_vert;
+
+			public TextToolButton (Gtk.Toolbar toolbar)
 			{
 				this.CanFocus = false;
 				this.Relief = Gtk.ReliefStyle.None;
 
-				Gtk.Image image = new Gtk.Image (Gtk.Stock.SelectFont, 
-								 Gtk.IconSize.LargeToolbar);
-				Gtk.Label label = new Gtk.Label (Catalog.GetString ("_Text"));
-				Gtk.Arrow arrow = new Gtk.Arrow (Gtk.ArrowType.Down, 
-								 Gtk.ShadowType.In);
+				image = new Gtk.Image (Gtk.Stock.SelectFont, toolbar.IconSize);
+				label_horiz = new Gtk.Label (Catalog.GetString ("_Text"));
+				label_vert = new Gtk.Label (Catalog.GetString ("_Text"));
+				arrow = new Gtk.Arrow (Gtk.ArrowType.Down, Gtk.ShadowType.In);
 
-				Gtk.HBox box = new Gtk.HBox (false, 4);
-				box.Add (image);
-				box.Add (label);
-				box.Add (arrow);
+				box_vert = new Gtk.VBox (false, 0);
+				box_vert.PackStart (image, true, true, 0);
+				box_vert.PackStart (label_vert, false, true, 0);
 
-				Gtk.Alignment align = new Gtk.Alignment (0.5f, 0.5f, 0.0f, 0.0f);
-				align.Add (box);
-				align.ShowAll ();
+				Gtk.HBox box_horiz = new Gtk.HBox (false, 0);
+				box_horiz.PackStart (box_vert, true, true, 0);
+				box_horiz.PackStart (label_horiz, true, true, 2);
+				box_horiz.PackStart (arrow, false, true, 0);
 
-				this.Add (align);
+				this.Add (box_horiz);
+				this.ShowAll ();
+				ShowForToolbarStyle (toolbar.ToolbarStyle);
+
+				toolbar.StyleChanged += OnToolbarStyleChanged;
 			}
 
 			protected override bool OnButtonPressEvent (Gdk.EventButton ev)
@@ -457,6 +518,37 @@ namespace Tomboy
 				base.OnButtonPressEvent (ev);
 				return false;
 			}
+
+			void ShowForToolbarStyle (Gtk.ToolbarStyle style)
+			{
+				switch (style) {
+				case Gtk.ToolbarStyle.Icons:
+					label_horiz.Hide ();
+					label_vert.Hide ();
+					image.Show ();
+					break;
+				case Gtk.ToolbarStyle.Text:
+					label_vert.Hide ();
+					image.Hide ();
+					label_horiz.Show ();
+					break;
+				case Gtk.ToolbarStyle.Both:
+					label_horiz.Hide ();
+					image.Show ();
+					label_vert.Show ();
+					break;
+				case Gtk.ToolbarStyle.BothHoriz:
+					label_vert.Hide ();
+					image.Show ();
+					label_horiz.Show ();
+					break;
+				}				
+			}
+
+			void OnToolbarStyleChanged (object sender, Gtk.StyleChangedArgs args)
+			{
+				ShowForToolbarStyle (args.Style);
+			}
 		}
 
 		// Find context menu
@@ -735,7 +827,7 @@ namespace Tomboy
 
 			huge = new Gtk.RadioMenuItem (normal.Group, 
 						      "<span size=\"x-large\">" +
-						      Catalog.GetString ("H_uge") +
+						      Catalog.GetString ("Hu_ge") +
 						      "</span>");
 			MarkupLabel (huge);
 			huge.Data ["Tag"] = "size:huge";
@@ -806,8 +898,6 @@ namespace Tomboy
 		{
 			event_freeze = true;
 
-			bool has_size = false;
-
 			bold.Active = buffer.IsActiveTag ("bold");
 			italic.Active = buffer.IsActiveTag ("italic");
 			strikeout.Active = buffer.IsActiveTag ("strikethrough");
