@@ -10,6 +10,10 @@ namespace Tomboy
 		static Gnome.Program program;
 		static Syscall.sighandler_t sig_handler;
 
+#if ENABLE_DBUS
+		static DBus.Connection dbus_connection;
+#endif
+
 		public static void Main (string [] args) 
 		{
 			RegisterSignalHandlers ();
@@ -30,10 +34,10 @@ namespace Tomboy
 						     Gnome.Modules.UI, 
 						     args);
 
-			if (cmd_line.UseTrayIcon) {
-				StartTrayIcon ();
-			} else {
+			if (cmd_line.UsePanelApplet) {
 				RegisterPanelAppletFactory (); 
+			} else {
+				StartTrayIcon ();
 			}
 
 			Console.WriteLine ("All done.  Ciao!");
@@ -57,9 +61,9 @@ namespace Tomboy
 		{
 #if ENABLE_DBUS
 			try {
-				DBus.Connection connection = DBus.Bus.GetSessionBus ();
+				dbus_connection = DBus.Bus.GetSessionBus ();
 				DBus.Service service = 
-					new DBus.Service (connection, 
+					new DBus.Service (dbus_connection, 
 							  RemoteControlProxy.Namespace);
 
 				RemoteControl remote_control = new RemoteControl (manager);
@@ -104,7 +108,7 @@ namespace Tomboy
 	public class TomboyCommandLine
 	{
 		bool new_note;
-		bool tray_icon;
+		bool panel_applet;
 		string new_note_name;
 		string open_note_uri;
 		string open_note_name;
@@ -115,9 +119,9 @@ namespace Tomboy
 			Parse (args);
 		}
 
-		public bool UseTrayIcon
+		public bool UsePanelApplet
 		{
-			get { return tray_icon; }
+			get { return panel_applet; }
 		}
 
 		public bool NeedsExecute
@@ -132,7 +136,7 @@ namespace Tomboy
                         string about = 
 				Catalog.GetString (
 					"Tomboy: A simple, easy to use desktop note-taking application.\n" +
-					"Copyright (C) 2004 Alex Graveley <alex@beatniksoftware.com>\n\n");
+					"Copyright (C) 2004, 2005 Alex Graveley <alex@beatniksoftware.com>\n\n");
 
 			Console.Write (about);
 		}
@@ -148,7 +152,6 @@ namespace Tomboy
 					"  --open-note [title/url]\tDisplay the existing note matching title.\n" +
 					"  --start-here\t\t\tDisplay the 'Start Here' note.\n" +
 					"  --highlight-search [text]\tSearch and highlight text in the opened note.\n" +
-					"  --tray-icon\t\t\tRun Tomboy in the system tray.\n" +
 					"  --version\t\t\tPrint version information.\n" +
 					"  --help\t\t\tPrint this usage message.\n");
 			Console.WriteLine (usage);
@@ -156,7 +159,6 @@ namespace Tomboy
 			string usage = 
 				Catalog.GetString (
 					"Usage:\n" +
-					"  --tray-icon\t\t\tRun Tomboy in the system tray.\n" +
 					"  --version\t\t\tPrint version information.\n" +
 					"  --help\t\t\tPrint this usage message.\n" +
 					"\n" +
@@ -232,8 +234,8 @@ namespace Tomboy
 					break;
 #endif // ENABLE_DBUS
 
-				case "--tray-icon":
-					tray_icon = true;
+				case "--panel-applet":
+					panel_applet = true;
 					break;
 
 				case "--version":
