@@ -1,125 +1,161 @@
 
 using System;
 using System.Collections;
+using System.IO;
 using System.Runtime.InteropServices;
 using System.Text.RegularExpressions;
 using Mono.Posix;
 
 namespace Tomboy
 {
-	public abstract class NotePlugin
-	{
-		Note note;
-
-		public void Initialize (Note note)
-		{
-			this.note = note;
-			Initialize ();
-		}
-
-		public abstract void Initialize ();
-
-		public Note Note
-		{
-			get { return note; }
-		}
-
-		public NoteBuffer Buffer
-		{
-			get { return note.Buffer; }
-		}
-
-		public NoteWindow Window
-		{
-			get { return note.Window; }
-		}
-
-		public NoteManager Manager
-		{
-			get { return note.Manager; }
-		}
-	}
-
 	public class NoteRenameWatcher : NotePlugin
 	{
-		Gtk.TextMark first_line_end;
-		uint rename_timeout_id;
+		bool editing_title;
 
-		public override void Initialize ()
+		protected override void Initialize ()
 		{
-			Note.Opened += OnNoteOpened;
+			// Do nothing.
 		}
 
-		void OnNoteOpened (object sender, EventArgs args)
+		protected override void Shutdown ()
 		{
+			// Do nothing.
+		}
+
+		Gtk.TextIter TitleEnd 
+		{
+			get {
+				Gtk.TextIter line_end = Buffer.StartIter;
+				line_end.ForwardToLineEnd ();
+				return line_end;
+			}
+		}
+
+		Gtk.TextIter TitleStart 
+		{
+			get { return Buffer.StartIter; }
+		}
+
+		protected override void OnNoteOpened ()
+		{
+			Buffer.MarkSet += OnMarkSet;
 			Buffer.InsertText += OnInsertText;
 			Buffer.DeleteRange += OnDeleteRange;
 
-			Gtk.TextIter line_end = Buffer.StartIter;
-			line_end.ForwardLine ();
+			// FIXME: Needed because we hide on delete event, and
+			// just hide on accelerator key, so we can't use delete
+			// event.  This means the window will flash if closed
+			// with a name clash.
+			Window.UnmapEvent += OnWindowClosed;
 
-			Console.WriteLine ("Applying <note-title> to '{0}'",
-					   Buffer.StartIter.GetText (line_end));
-
-			Buffer.ApplyTag ("note-title", Buffer.StartIter, line_end);
-
-			first_line_end = Buffer.CreateMark (null, 
-							    line_end, 
-							    false /* keep to the right */);
+			// Clean up title line
+			Buffer.RemoveAllTags (TitleStart, TitleEnd);
+			Buffer.ApplyTag ("note-title", TitleStart, TitleEnd);
 		}
 
-		void OnDeleteRange (object sender, Gtk.DeleteRangeArgs args)
+		// This only gets called on an explicit move, not when typing
+		void OnMarkSet (object sender, Gtk.MarkSetArgs args)
 		{
-			if (args.Start.Line != 0)
-				return;
-
-			UpdateTitle ();
+			Update ();
 		}
 
 		void OnInsertText (object sender, Gtk.InsertTextArgs args)
 		{
-			if (args.Pos.Line != 0)
-				return;
+			Changed ();
+			Update ();
 
-			UpdateTitle ();
+			Gtk.TextIter end = args.Pos;
+			end.ForwardToLineEnd ();
+
+			// Avoid lingering note-title after a multi-line insert...
+			Buffer.RemoveTag ("note-title", TitleEnd, end);
 		}
 
-		void UpdateTitle ()
+		void OnDeleteRange (object sender, Gtk.DeleteRangeArgs args)
 		{
-			// Replace the existing save timeout...
-			if (rename_timeout_id != 0)
-				GLib.Source.Remove (rename_timeout_id);
+			Changed ();
+			Update ();
+		}
 
-			// Wait .5 seconds before renaming...
-			rename_timeout_id = 
-				GLib.Timeout.Add (500, 
-						  new GLib.TimeoutHandler (UpdateTitleTimeout));
+		void Update ()
+		{
+			Gtk.TextIter insert = Buffer.GetIterAtMark (Buffer.InsertMark);
+			Gtk.TextIter selection = Buffer.GetIterAtMark (Buffer.SelectionBound);
 
-			Gtk.TextIter line_end = Buffer.GetIterAtMark (first_line_end);
-			Buffer.ApplyTag ("note-title", Buffer.StartIter, line_end);
+			if (insert.Line == 0 || selection.Line == 0) {
+				if (!editing_title)
+					editing_title = true;
+			} else {
+				if (editing_title) {
+					UpdateNoteTitle ();
+					editing_title = false;
+				}
+			}
+		}
+
+		void Changed ()
+		{
+			if (!editing_title)
+				return;
+
+			// Make sure the title line is big and red...
+			Buffer.RemoveAllTags (TitleStart, TitleEnd);
+			Buffer.ApplyTag ("note-title", TitleStart, TitleEnd);
+
+			// NOTE: Use "(Untitled #)" for empty first lines...
+			string title = TitleStart.GetText (TitleEnd).Trim ();
+			if (title == string.Empty)
+				title = GetUniqueUntitled ();
 
 			// Only set window title here, to give feedback that we
 			// are indeed changing the title.
-			Window.Title = Buffer.StartIter.GetText (line_end).Trim ();
+			Window.Title = title;
 		}
 
-		bool UpdateTitleTimeout ()
+		[GLib.ConnectBefore]
+		void OnWindowClosed (object sender, Gtk.UnmapEventArgs args)
 		{
-			Gtk.TextIter line_end = Buffer.GetIterAtMark (first_line_end);
-			string title = Buffer.StartIter.GetText (line_end).Trim ();
+			if (!editing_title)
+				return;
+
+			if (!UpdateNoteTitle ()) {
+				args.RetVal = true;
+				return;
+			}
+		}
+
+		string GetUniqueUntitled ()
+		{
+			int new_num = Manager.Notes.Count;
+			string temp_title;
+
+			while (true) {
+				temp_title = String.Format (Catalog.GetString ("(Untitled {0})"), 
+							    ++new_num);
+				if (Manager.Find (temp_title) == null)
+					return temp_title;
+			}
+		}
+
+		bool UpdateNoteTitle ()
+		{
+			string title = Window.Title;
 
 			Note existing = Manager.Find (title);
-
 			if (existing != null && existing != this.Note) {
-				ShowNameClashError (Note, title);
-			} else {
-				Note.Title = title;
+				// Present the window in case it got unmapped...
+				// FIXME: Causes flicker.
+				Note.Window.Present ();
+
+				ShowNameClashError (title);
+				return false;
 			}
 
-			return false;
+			Note.Title = title;
+			return true;
 		}
 
-		void ShowNameClashError (Note note, string title)
+		void ShowNameClashError (string title)
 		{
 			string message = 
 				String.Format (Catalog.GetString ("A note with the title " +
@@ -142,54 +178,239 @@ namespace Tomboy
 		}
 	}
 
+	public class NoteRelatedToWatcher : NotePlugin
+	{
+		ArrayList linked_from;
+		ArrayList linked_to;
+
+		protected override void Initialize ()
+		{
+			linked_from = new ArrayList ();
+			linked_to = new ArrayList ();
+
+			LinkCreated += OnLinkCreated;
+		}
+
+		protected override void Shutdown ()
+		{
+			// Do nothing.
+		}
+
+		protected override void OnNoteOpened ()
+		{
+			Buffer.TagApplied += OnTagApplied;
+		}
+
+		void OnTagApplied (object sender, Gtk.TagAppliedArgs args)
+		{
+			if (args.Tag.Name != "link:internal") 
+				return;
+
+			string link_text = args.StartChar.GetText (args.EndChar);
+
+			// Ignore self-references
+			if (link_text.ToLower () == Note.Title.ToLower ())
+				return;
+
+			linked_to.Add (link_text);
+
+			Console.WriteLine ("Link Tag applied to '{0}', for '{1}'",
+					   Note.Title,
+					   link_text);
+
+			LinkCreated (Note, link_text);
+		}
+
+		void OnTagRemoved (object sender, Gtk.TagAppliedArgs args)
+		{
+			if (args.Tag.Name != "link:internal")
+				return;
+
+			Console.WriteLine ("Tag 'link:internal' removed!!!");
+		}
+
+		void OnLinkCreated (Note note, string link_text)
+		{
+			if (link_text.ToLower () != Note.Title.ToLower ())
+				return;
+
+			Console.WriteLine ("Received Link text from '{0}' = '{1}'",
+					   note.Title,
+					   link_text);
+
+			// Don't show links we already link to in the body
+			if (linked_to.Contains (link_text))
+				return;
+
+			if (linked_from.Contains (note.Title))
+				return;
+
+			linked_from.Add (note.Title);
+
+			Gtk.TextTag tag = Buffer.TagTable.Lookup ("related-to");
+			Gtk.TextIter iter = Buffer.StartIter;
+
+			if (iter.ForwardToTagToggle (tag)) 
+				RemoveRelatedLine ();
+
+			InsertRelatedLine ();
+		}
+
+		void InsertRelatedLine ()
+		{
+			Gtk.TextIter line, line_end;
+
+			line = Buffer.StartIter;
+			line.ForwardLines (1);
+
+			Buffer.Insert (line, Catalog.GetString ("Related to: "));
+
+			line = Buffer.StartIter;
+			line.ForwardLines (1);
+			line_end = line;
+			line_end.ForwardToLineEnd ();
+
+			Buffer.ApplyTag ("italic", line, line_end);
+
+			line = Buffer.StartIter;
+			line.ForwardLines (1);
+			line_end = line;
+			line_end.ForwardToLineEnd ();
+
+			string link_text = string.Empty;
+			foreach (string link in linked_from) {
+				if (link_text == string.Empty)
+					link_text = link;
+				else {
+					link_text += ", ";
+					link_text += link;
+				}
+			}
+			link_text += "\n";
+
+			Buffer.Insert (line_end, link_text);
+
+			line = Buffer.StartIter;
+			line.ForwardLines (1);
+			line_end = line;
+			line_end.ForwardToLineEnd ();
+
+			Buffer.ApplyTag ("related-to", line, line_end);
+		}
+
+		void RemoveRelatedLine ()
+		{
+			Gtk.TextTag tag = Buffer.TagTable.Lookup ("related-to");
+
+			Gtk.TextIter line = Buffer.StartIter;
+			line.ForwardToTagToggle (tag);
+
+			Gtk.TextIter line_end = line;
+			line_end.ForwardToTagToggle (tag);
+			
+			Buffer.Delete (line, line_end);
+		}
+
+		delegate void LinkCreatedHandler (Note note, string linked_text);
+
+		static event LinkCreatedHandler LinkCreated;
+	}
+
 	public class NoteSpellChecker : NotePlugin
 	{
-		[DllImport ("libgtkspell.so.0")]
-		static extern void gtkspell_new_attach (IntPtr text_view, 
-							string locale, 
-							IntPtr error);
+		IntPtr obj_ptr = IntPtr.Zero;
 
-		public override void Initialize ()
+		[DllImport ("libgtkspell")]
+		static extern IntPtr gtkspell_new_attach (IntPtr text_view, 
+							  string locale, 
+							  IntPtr error);
+
+		[DllImport ("libgtkspell")]
+		static extern void gtkspell_detach (IntPtr obj);
+
+		protected override void Initialize ()
 		{
-			Note.Opened += OnNoteOpened;
+			// Do nothing.
 		}
 
-		public void OnNoteOpened (object sender, EventArgs args)
+		protected override void Shutdown ()
+		{
+			// Do nothing.
+		}
+
+		protected override void OnNoteOpened ()
 		{
 			Buffer.TagApplied += TagApplied;
+			Preferences.SettingChanged += OnEnableSpellcheckChanged;
 
-			gtkspell_new_attach (Window.Editor.Handle, 
-					     null, 
-					     IntPtr.Zero);
-
-			FixupOldGtkSpell ();
+			if ((bool) Preferences.Get (Preferences.ENABLE_SPELLCHECKING)) {
+				obj_ptr = gtkspell_new_attach (Window.Editor.Handle, 
+							       null, 
+							       IntPtr.Zero);
+				FixupOldGtkSpell ();
+			}
 		}
 
-		[System.Diagnostics.Conditional ("OLD_GTKSPELL")] 
-		void FixupOldGtkSpell ()
+		// To unset GtkSpell's tag Foreground correctly we need to call
+		// SetProperty, which is a protected member.
+		class OldGtkSpellFixup : Gtk.TextTag
 		{
-			// NOTE: Older versions of GtkSpell before 2.0.6 use red
-			// foreground color and a single underline.  This
-			// conflicts with internal note links.  So fix it up to
-			// use the "normal" foreground and the "error"
-			// underline.
-			Gtk.TextTag misspell = Buffer.TagTable.Lookup ("gtkspell-misspelled");
+			public OldGtkSpellFixup (IntPtr raw)
+				: base (raw)
+			{
+			}
+
+			public void Fixup ()
+			{
+				SetProperty ("foreground_gdk", new GLib.Value (Gdk.Color.Zero));
 
-			if (misspell != null) {
-				Gtk.TextTag normal = Buffer.TagTable.Lookup ("normal");
-				misspell.ForegroundGdk = normal.ForegroundGdk;
 				// Force the value to 4 since error underlining
 				// isn't mapped in Gtk# yet.
-				misspell.Underline = (Pango.Underline) 4;
+				Underline = (Pango.Underline) 4;
+			}
+		}
+
+		// NOTE: Older versions of GtkSpell before 2.0.6 use red
+		// foreground color and a single underline.  This conflicts with
+		// internal note links.  So fix it up to unset foreground and
+		// use the "error" underline.
+		void FixupOldGtkSpell ()
+		{
+			Gtk.TextTag misspell = Buffer.TagTable.Lookup ("gtkspell-misspelled");
+			if (misspell != null) {
+				OldGtkSpellFixup hack = new OldGtkSpellFixup (misspell.Handle);
+				hack.Fixup ();
+			}
+		}
+
+		void OnEnableSpellcheckChanged (object sender, GConf.NotifyEventArgs args)
+		{
+			if (args.Key != Preferences.ENABLE_SPELLCHECKING)
+				return;
+
+			if ((bool) args.Value) {
+				if (obj_ptr == IntPtr.Zero) {
+					obj_ptr = gtkspell_new_attach (Window.Editor.Handle, 
+								       null, 
+								       IntPtr.Zero);
+					FixupOldGtkSpell ();
+				}
+			} else {
+				if (obj_ptr != IntPtr.Zero) {
+					gtkspell_detach (obj_ptr);
+					obj_ptr = IntPtr.Zero;
+				}
 			}
 		}
 
 		void TagApplied (object sender, Gtk.TagAppliedArgs args)
 		{
 			if (args.Tag.Name == "gtkspell-misspelled") {
-				// Remove misspelled tag for links 
+				// Remove misspelled tag for links & title
 				foreach (Gtk.TextTag tag in args.StartChar.Tags) {
-					if (tag.Name != null && tag.Name.StartsWith ("link:")) {
+					if (tag.Name != null && 
+					    (tag.Name.StartsWith ("link:") || 
+					     tag.Name == "note-title")) {
 						Buffer.RemoveTag ("gtkspell-misspelled", 
 								  args.StartChar, 
 								  args.EndChar);
@@ -197,10 +418,14 @@ namespace Tomboy
 					}
 				}
 			} else if (args.Tag.Name != null && 
-				   args.Tag.Name.StartsWith ("link:")) {
-				Buffer.RemoveTag ("gtkspell-misspelled", 
-						  args.StartChar, 
-						  args.EndChar);
+				   (args.Tag.Name.StartsWith ("link:") ||
+				    args.Tag.Name == "note-title")) {
+				Gtk.TextTag misspell = Buffer.TagTable.Lookup ("gtkspell-misspelled");
+				if (misspell != null) {
+					Buffer.RemoveTag ("gtkspell-misspelled", 
+							  args.StartChar, 
+							  args.EndChar);
+				}
 			}
 		}
 	}
@@ -209,11 +434,8 @@ namespace Tomboy
 	{
 		Gtk.TextTag url_tag;
 
-		// FIXME: Bugs because of word boundries.
-		// 	- Requires a character after www. or http:// to be typed to hightlight
-		//	- doesn't match /unix/path (only /third/unix/path)
 		const string URL_REGEX = 
-			@"\b(((news|http|https|ftp|irc)://|(mailto:)|((www|ftp)\.)|/\S+/|\S*@\S*\.)(\S*[^\.,!\?;:])?)\b";
+			@"((\b((news|http|https|ftp|file|irc)://|mailto:|(www|ftp)\.|\S*@\S*\.)|(^|\s)/\S+/|(^|\s)~/\S+)\S*\b/?)";
 
 		static Regex regex;
 
@@ -223,12 +445,17 @@ namespace Tomboy
 					   RegexOptions.IgnoreCase | RegexOptions.Compiled);
 		}
 
-		public override void Initialize ()
+		protected override void Initialize ()
 		{
-			Note.Opened += OnNoteOpened;
+			// Do nothing.
 		}
 
-		public void OnNoteOpened (object sender, EventArgs args)
+		protected override void Shutdown ()
+		{
+			// Do nothing.
+		}
+
+		protected override void OnNoteOpened ()
 		{
 			url_tag = Buffer.TagTable.Lookup ("link:url");
 			if (url_tag == null) {
@@ -239,7 +466,7 @@ namespace Tomboy
 			url_tag.TextEvent += OnTextEvent;
 
 			Buffer.InsertText += OnInsertText;
-			Buffer.DeleteRange += OnDeleteRange;			
+			Buffer.DeleteRange += OnDeleteRange;
 		}
 
 		void OnTextEvent (object sender, Gtk.TextEventArgs args)
@@ -258,13 +485,18 @@ namespace Tomboy
 						      Gdk.ModifierType.ControlMask)) != 0)
 				return;
 
-			Gtk.TextIter start = args.Iter, end = args.Iter;
-
+			Gtk.TextIter start = args.Iter;
 			start.BackwardToTagToggle (tag);
+
+			Gtk.TextIter end = args.Iter;
 			end.ForwardToTagToggle (tag);
 
 			string url = start.GetText (end);
 
+			// FIXME: Needed because the file match is greedy and
+			// eats a leading space.
+			url = url.Trim ();
+
 			// Simple url massaging.  Add to 'http://' to the front
 			// of www.foo.com, 'mailto:' to alex@foo.com, 'file://'
 			// to /home/alex/foo.
@@ -278,6 +510,10 @@ namespace Tomboy
 			else if (url.StartsWith ("/") && 
 				 url.LastIndexOf ("/") > 1)
 				url = "file://" + url;
+			else if (url.StartsWith ("~/"))
+				url = "file://" + 
+					Path.Combine (Environment.GetEnvironmentVariable ("HOME"),
+						      url.Substring (2));
 			else if (url.IndexOf ("@") > 1 &&
 				 url.IndexOf (".") > 3 &&
 				 !url.StartsWith ("mailto:"))
@@ -288,7 +524,7 @@ namespace Tomboy
 			try {
 				Gnome.Url.Show (url);
 
-				/* Close note on middle-click */
+				// Close note on middle-click
 				if (button_ev.Button == 2) {
 					Window.Hide ();
 
@@ -330,9 +566,11 @@ namespace Tomboy
 			     match = match.NextMatch ()) {
 				Group group = match.Groups [1];
 
+				/*
 				Console.WriteLine("Highlighting url: '{0}' at offset {1}",
 						  group,
 						  group.Index);
+				*/
 
 				Gtk.TextIter start_cpy = start;
 				start_cpy.ForwardChars (group.Index);
@@ -360,210 +598,249 @@ namespace Tomboy
 
 	public class NoteLinkWatcher : NotePlugin
 	{
-		ArrayList note_titles;
-		int longest_title;
-
-		public override void Initialize () 
+		class TrieController
 		{
-			Note.Opened += OnNoteOpened;
+			TrieTree title_trie;
+			NoteManager manager;
 
+			public TrieController (NoteManager manager)
+			{
+				this.manager = manager;
+				manager.NoteDeleted += OnNoteDeleted;
+				manager.NoteAdded += OnNoteAdded;
+				manager.NoteRenamed += OnNoteRenamed;
+
+				Update ();
+			}
+
+			void OnNoteAdded (object sender, Note added)
+			{
+				Update ();
+			}
+
+			void OnNoteDeleted (object sender, Note deleted)
+			{
+				Update ();
+			}
+
+			void OnNoteRenamed (Note renamed, string old_title)
+			{
+				Update ();
+			}
+
+			public void Update ()
+			{
+				title_trie = new TrieTree (false /* !case_sensitive */);
+
+				foreach (Note note in manager.Notes) {
+					title_trie.AddKeyword (note.Title, note);
+				}
+			}
+
+			public TrieTree TitleTrie 
+			{
+				get { return title_trie; }
+			}
+		}
+
+		static TrieController trie_controller;
+
+		protected override void Initialize () 
+		{
 			Manager.NoteDeleted += OnNoteDeleted;
 			Manager.NoteAdded += OnNoteAdded;
 			Manager.NoteRenamed += OnNoteRenamed;
-
-			note_titles = new ArrayList ();
 		}
 
-		void OnNoteOpened (object sender, EventArgs args)
+		protected override void Shutdown ()
+		{
+			Manager.NoteDeleted -= OnNoteDeleted;
+			Manager.NoteAdded -= OnNoteAdded;
+			Manager.NoteRenamed -= OnNoteRenamed;
+		}
+
+		protected override void OnNoteOpened ()
 		{
 			Buffer.InsertText += OnInsertText;
 			Buffer.DeleteRange += OnDeleteRange;
-
-			UpdateTitleCache ();
-
-			// Do we need this?  Causes unnecessary write on open.
-			//HighlightInBlock (Buffer.Text.ToLower (), Buffer.StartIter);
-		}
-
-		// Updating this on NoteChanged allows us to fetch a smaller
-		// block to look for links when typing, and avoid lowercasing...
-		void UpdateTitleCache ()
-		{
-			longest_title = 0;
-			note_titles.Clear ();
-
-			foreach (Note note in Manager.Notes) {
-				if (note.Title.Length > longest_title)
-					longest_title = note.Title.Length;
-
-				if (note.Title.Length > 0)
-					note_titles.Add (note.Title.ToLower ());
-			}
-		}
-
-		void ReplaceTagOnMatch (Gtk.TextTag find_tag,
-					Gtk.TextTag replace_with_tag,
-					string      match)
-		{
-			Gtk.TextIter iter = Buffer.StartIter;
-
-			while (iter.ForwardToTagToggle (find_tag) &&
-			       !iter.Equal (Buffer.EndIter)) {
-				if (!iter.BeginsTag (find_tag))
-					continue;
-
-				Gtk.TextIter end = iter;
-				if (!end.ForwardToTagToggle (find_tag))
-					break;
-
-				string tag_content = iter.GetText (end).ToLower ();
-
-				if (tag_content == match) {
-					Buffer.RemoveTag (find_tag, iter, end);
-					Buffer.ApplyTag (replace_with_tag, iter, end);
-				}
-
-				iter = end;
-			}
 		}
 
 		bool ContainsText (string text)
 		{
-			// FIXME: XML Encode text, as Note.Text is encoded.
-			return Note.Text.ToLower ().IndexOf (text.ToLower ()) > -1;
+			string body = Note.TextContent.ToLower ();
+			string match = text.ToLower ();
+
+			return body.IndexOf (match) > -1;
 		}
 
-		// This can be called whether Note is showing or not.  If
-		// showing, manually apply the link tags to Buffer.  Otherwise
-		// just replace the xml Note.Text and queue a save.
 		void OnNoteAdded (object sender, Note added)
 		{
-			UpdateTitleCache ();
-
 			if (added == this.Note)
 				return;
 
-			if (ContainsText (added.Title)) {
-				HighlightInBlock (Buffer.Text.ToLower (), Buffer.StartIter);
-			}
-		}
-
-		// This can be called whether Note is showing or not.  If
-		// showing, manually apply the link tags to Buffer.  Otherwise
-		// just replace the xml Note.Text and queue a save.
-		void OnNoteDeleted (object sender, Note deleted)
-		{
-			UpdateTitleCache ();
-
-			if (deleted == this.Note)
+			if (!ContainsText (added.Title))
 				return;
 
-			if (ContainsText (deleted.Title)) {
-				Gtk.TextTag link_tag = Buffer.TagTable.Lookup ("link:internal");
-				Gtk.TextTag broken_link_tag = Buffer.TagTable.Lookup ("link:broken");
-
-				// Switch any link:internal tags which point to
-				// the deleted note to link:broken
-				ReplaceTagOnMatch (link_tag,
-						   broken_link_tag,
-						   deleted.Title.ToLower ());
-			}
-		}
-
-		// This can be called whether Note is showing or not.  If
-		// showing, manually change the Buffer content.  Otherwise
-		// just replace the xml Note.Text and queue a save.
-		void OnNoteRenamed (Note renamed, string old_title)
-		{
-			UpdateTitleCache ();
-
-			if (ContainsText (old_title)) {
-				Gtk.TextTag link_tag = Buffer.TagTable.Lookup ("link:internal");
-				
-				Gtk.TextIter iter = Buffer.StartIter;
-				string old_title_lower = old_title.ToLower ();
-
-				while (iter.ForwardToTagToggle (link_tag) &&
-				       !iter.Equal (Buffer.EndIter)) {
-					if (!iter.BeginsTag (link_tag))
-						continue;
-
-					Gtk.TextIter end = iter;
-					if (!end.ForwardToTagToggle (link_tag))
-						break;
-
-					string tag_content = iter.GetText (end).ToLower ();
-
-					if (tag_content == old_title_lower) {
-						Console.WriteLine ("Replacing with '{0}'", renamed.Title);
-
-						int iter_offset = iter.Offset;
-						int end_offset = end.Offset;
-
-						Buffer.Delete (iter, end);
-
-						iter = Buffer.GetIterAtOffset (iter_offset);
-						end = Buffer.GetIterAtOffset (end_offset);
-
-						Buffer.Insert (iter, renamed.Title);
-
-						iter = Buffer.GetIterAtOffset (iter_offset);
-						end = Buffer.GetIterAtOffset (end_offset);
-					}
-
-					iter = end;
-				}
-			}
-		}
-
-		void HighlightInBlock (string lower_text, Gtk.TextIter cursor) 
-		{
+			Gtk.TextTag url_tag = Buffer.TagTable.Lookup ("link:url");
 			Gtk.TextTag link_tag = Buffer.TagTable.Lookup ("link:internal");
 			Gtk.TextTag broken_link_tag = Buffer.TagTable.Lookup ("link:broken");
 
-			foreach (string title in note_titles) {
-				int last_idx = 0;
+			string new_title_lower = added.Title.ToLower ();
 
-				if (title == Note.Title)
-					continue;
+			string buffer_text = Buffer.StartIter.GetText (Buffer.EndIter);
+			buffer_text = buffer_text.ToLower ();
 
-				while (true) {
-					int idx = lower_text.IndexOf (title, last_idx);
-					if (idx < 0)
-						break;
+			int idx = 0;
 
-					Gtk.TextIter title_start = cursor;
-					title_start.ForwardChars (idx);
+			while (true) {
+				idx = buffer_text.IndexOf (new_title_lower, idx);
+				if (idx < 0)
+					break;
 
-					Gtk.TextIter title_end = title_start;
-					title_end.ForwardChars (title.Length);
+				Gtk.TextIter start = Buffer.GetIterAtOffset (idx);
+				Gtk.TextIter end = Buffer.GetIterAtOffset (idx + added.Title.Length);
 
-					Console.WriteLine ("Matching Note title '{0}'...", title);
-
-					Buffer.RemoveTag (broken_link_tag,
-							  title_start,
-							  title_end);
-					Buffer.ApplyTag (link_tag,
-							 title_start,
-							 title_end);
-
-					last_idx = idx + title.Length;
+				if (!start.HasTag (url_tag)) {
+					Buffer.RemoveTag (broken_link_tag, start, end);
+					Buffer.ApplyTag (link_tag, start, end);
 				}
+
+				idx += new_title_lower.Length;
 			}
 		}
 
-		void UnhighlightInBlock (string text, Gtk.TextIter cursor, Gtk.TextIter end) 
+		void OnNoteDeleted (object sender, Note deleted)
 		{
+			if (deleted == this.Note)
+				return;
+
+			if (!ContainsText (deleted.Title))
+				return;
+
 			Gtk.TextTag link_tag = Buffer.TagTable.Lookup ("link:internal");
-			Buffer.RemoveTag (link_tag, cursor, end);
+			Gtk.TextTag broken_link_tag = Buffer.TagTable.Lookup ("link:broken");
+
+			string old_title_lower = deleted.Title.ToLower ();
+
+			TextTagEnumerator enumerator = new TextTagEnumerator (Buffer, link_tag);
+			foreach (TextRange range in enumerator) {
+				if (range.Text.ToLower () != old_title_lower)
+					continue;
+
+				Buffer.RemoveTag (link_tag, range.Start, range.End);
+				Buffer.ApplyTag (broken_link_tag, range.Start, range.End);
+			}
 		}
 
-		string GetBlockAroundCursor (ref Gtk.TextIter start, ref Gtk.TextIter end) 
+		void OnNoteRenamed (Note renamed, string old_title)
 		{
+			if (!ContainsText (old_title))
+				return;
+
+			Gtk.TextTag link_tag = Buffer.TagTable.Lookup ("link:internal");
+
+			string old_title_lower = old_title.ToLower ();
+
+			TextTagEnumerator enumerator = new TextTagEnumerator (Buffer, link_tag);
+			foreach (TextRange range in enumerator) {
+				Console.WriteLine ("RENAME: Checking '{0}' in note '{1}'", range.Text, Note.Title);
+
+				if (range.Text.ToLower () != old_title_lower)
+					continue;
+
+				Console.WriteLine ("Replacing with '{0}'", renamed.Title);
+
+				Buffer.Delete (range.Start, range.End);
+				Buffer.InsertWithTags (range.Start, renamed.Title, link_tag);
+			}
+		}
+
+		class Highlighter
+		{
+			Note note;
+			Gtk.TextIter start;
+			Gtk.TextIter end;
+
+			Gtk.TextTag url_tag;
+			Gtk.TextTag link_tag;
+			Gtk.TextTag broken_link_tag;
+
+			public Highlighter (Note         note, 
+					    Gtk.TextIter start, 
+					    Gtk.TextIter end)
+			{
+				this.note = note;
+				this.start = start;
+				this.end = end;
+
+				this.url_tag = note.Buffer.TagTable.Lookup ("link:url");
+				this.link_tag = note.Buffer.TagTable.Lookup ("link:internal");
+				this.broken_link_tag = note.Buffer.TagTable.Lookup ("link:broken");
+			}
+
+			public void Highlight (TrieTree trie)
+			{
+				trie.FindMatches (start.GetText (end), 
+						  new MatchHandler (TitleFound));
+			}
+
+			void TitleFound (string haystack, int start_idx, object value)
+			{
+				Note match_note = (Note) value;
+
+				if (match_note == note)
+					return;
+
+				Gtk.TextIter title_start = start;
+				title_start.ForwardChars (start_idx);
+
+				// Don't create links inside URLs
+				if (title_start.HasTag (url_tag))
+					return;
+
+				Gtk.TextIter title_end = title_start;
+				title_end.ForwardChars (match_note.Title.Length);
+
+				Console.WriteLine ("Matching Note title '{0}'...", match_note.Title);
+
+				note.Buffer.RemoveTag (broken_link_tag,
+						       title_start,
+						       title_end);
+				note.Buffer.ApplyTag (link_tag,
+						      title_start,
+						      title_end);
+			}
+		}
+
+		TrieTree TitleTrie
+		{
+			get {
+				if (trie_controller == null)
+					trie_controller = new TrieController (Manager);
+
+				return trie_controller.TitleTrie;
+			}
+		}
+
+		void HighlightInBlock (Gtk.TextIter start, Gtk.TextIter end) 
+		{
+			Highlighter high = new Highlighter (Note, start, end);
+			high.Highlight (TitleTrie);
+		}
+
+		void UnhighlightInBlock (Gtk.TextIter start, Gtk.TextIter end) 
+		{
+			Gtk.TextTag link_tag = Buffer.TagTable.Lookup ("link:internal");
+			Buffer.RemoveTag (link_tag, start, end);
+		}
+
+		void GetBlockExtents (ref Gtk.TextIter start, ref Gtk.TextIter end) 
+		{
+			// FIXME: Should only be processing the largest title
+			// size, so we don't slow down for large paragraphs 
+
 			start.LineOffset = 0;
 			end.ForwardToLineEnd ();
-
-			return start.GetText (end).ToLower ();
 		}
 
 		void OnDeleteRange (object sender, Gtk.DeleteRangeArgs args)
@@ -571,14 +848,10 @@ namespace Tomboy
 			Gtk.TextIter start = args.Start;
 			Gtk.TextIter end = args.End;
 
-			// Avoid title line
-			if (start.Line == 0)
-				return;
+			GetBlockExtents (ref start, ref end);
 
-			string block = GetBlockAroundCursor (ref start, ref end);
-
-			UnhighlightInBlock (block, start, end);
-			HighlightInBlock (block, start);
+			UnhighlightInBlock (start, end);
+			HighlightInBlock (start, end);
 		}
 
 		void OnInsertText (object sender, Gtk.InsertTextArgs args)
@@ -588,14 +861,10 @@ namespace Tomboy
 
 			Gtk.TextIter end = args.Pos;
 
-			// Avoid title line
-			if (start.Line == 0)
-				return;
+			GetBlockExtents (ref start, ref end);
 
-			string block = GetBlockAroundCursor (ref start, ref end);
-
-			UnhighlightInBlock (block, start, end);
-			HighlightInBlock (block, start);
+			UnhighlightInBlock (start, end);
+			HighlightInBlock (start, end);
 		}
 	}
 
@@ -603,7 +872,8 @@ namespace Tomboy
 	{
 		Gtk.TextTag broken_link_tag;
 
-		const string WIKIWORD_REGEX = @"\b(([A-Z]+[a-z0-9]+){2}([A-Za-z0-9])*)\b";
+		// NOTE: \p{Lu} is Unicode uppercase and \p{Ll} is lowercase.
+		const string WIKIWORD_REGEX = @"\b((\p{Lu}+[\p{Ll}0-9]+){2}([\p{Lu}\p{Ll}0-9])*)\b";
 
 		static Regex regex; 
 
@@ -612,12 +882,17 @@ namespace Tomboy
 			regex = new Regex (WIKIWORD_REGEX, RegexOptions.Compiled);
 		}
 
-		public override void Initialize ()
+		protected override void Initialize ()
 		{
-			Note.Opened += OnNoteOpened;
+			// Do nothing.
 		}
 
-		void OnNoteOpened (object sender, EventArgs args)
+		protected override void Shutdown ()
+		{
+			// Do nothing.
+		}
+
+		protected override void OnNoteOpened ()
 		{
 			broken_link_tag = Buffer.TagTable.Lookup ("link:broken");
 			if (broken_link_tag == null) {
@@ -625,8 +900,39 @@ namespace Tomboy
 				return;
 			}
 
-			Buffer.InsertText += OnInsertText;
-			Buffer.DeleteRange += OnDeleteRange;
+			if ((bool) Preferences.Get (Preferences.ENABLE_WIKIWORDS)) {
+				Buffer.InsertText += OnInsertText;
+				Buffer.DeleteRange += OnDeleteRange;
+			}
+			Preferences.SettingChanged += OnEnableWikiwordsChanged;
+		}
+
+		void OnEnableWikiwordsChanged (object sender, GConf.NotifyEventArgs args)
+		{
+			if (args.Key != Preferences.ENABLE_WIKIWORDS)
+				return;
+
+			if ((bool) args.Value) {
+				Buffer.InsertText += OnInsertText;
+				Buffer.DeleteRange += OnDeleteRange;
+			} else {
+				Buffer.InsertText -= OnInsertText;
+				Buffer.DeleteRange -= OnDeleteRange;
+			}
+		}
+
+		static string [] PatronymicPrefixes = 
+			new string [] { "Mc", "Mac", "Le", "La", "De", "Van" };
+
+		bool IsPatronymicName (string word)
+		{
+			foreach (string prefix in PatronymicPrefixes) {
+				if (word.StartsWith (prefix) &&
+				    char.IsUpper (word [prefix.Length]))
+					return true;
+			}
+
+			return false;
 		}
 
 		void ApplyWikiwordToBlock (Gtk.TextIter start, Gtk.TextIter end)
@@ -641,6 +947,9 @@ namespace Tomboy
 			     match = match.NextMatch ()) {
 				Group group = match.Groups [1];
 
+				if (IsPatronymicName (group.ToString ()))
+					continue;
+
 				Console.WriteLine("Highlighting wikiword: '{0}' at offset {1}",
 						  group,
 						  group.Index);
@@ -687,12 +996,17 @@ namespace Tomboy
 			hand_cursor = new Gdk.Cursor (Gdk.CursorType.Hand2);
 		}
 
-		public override void Initialize ()
+		protected override void Initialize ()
 		{
-			Note.Opened += OnNoteOpened;
+			// Do nothing.
 		}
 
-		public void OnNoteOpened (object sender, EventArgs args)
+		protected override void Shutdown ()
+		{
+			// Do nothing.
+		}
+
+		protected override void OnNoteOpened ()
 		{
 			link_tag = Buffer.TagTable.Lookup ("link:internal");
 			link_tag.TextEvent += OnLinkTextEvent;
@@ -743,7 +1057,8 @@ namespace Tomboy
 
 			Gtk.TextIter start = args.Iter, end = args.Iter;
 
-			start.BackwardToTagToggle (tag);
+			if (!start.BeginsTag (tag))
+				start.BackwardToTagToggle (tag);
 			end.ForwardToTagToggle (tag);
 
 			if (!OpenOrCreateLink (start, end))
@@ -791,7 +1106,8 @@ namespace Tomboy
 					if (tag == link_tag || tag == broken_link_tag) {
 						Gtk.TextIter start = iter, end = iter;
 
-						start.BackwardToTagToggle (tag);
+						if (!start.BeginsTag (tag))
+							start.BackwardToTagToggle (tag);
 						end.ForwardToTagToggle (tag);
 
 						OpenOrCreateLink (start, end);
@@ -871,21 +1187,50 @@ namespace Tomboy
 
 	public class IndentWatcher : NotePlugin
 	{
-		public override void Initialize ()
+		protected override void Initialize ()
 		{
-			Note.Opened += OnNoteOpened;
+			// Do nothing.
 		}
 
-		public void OnNoteOpened (object sender, EventArgs args)
+		protected override void Shutdown ()
 		{
-			Gtk.TextView editor = Window.Editor;
-			editor.KeyPressEvent += OnEditorKeyPress;
+			// Do nothing.
+		}
+
+		protected override void OnNoteOpened ()
+		{
+			Window.Editor.KeyPressEvent += OnEditorKeyPress;
+			Buffer.InsertText += OnInsertText;
 		}
 
 		[GLib.ConnectBefore]
 		void OnEditorKeyPress (object sender, Gtk.KeyPressEventArgs args) 
 		{
 			switch (args.Event.Key) {
+				/*
+			case Gdk.Key.asterisk:
+				Console.WriteLine ("Got Asterisk!");
+
+				{
+				Gtk.TextIter insert = Buffer.GetIterAtMark (Buffer.InsertMark);
+				if (!insert.StartsLine ())
+					break;
+
+				Gtk.TextIter para_end = insert;
+				para_end.ForwardToLineEnd ();
+
+				Console.WriteLine ("Setting hanging indent!");
+
+				Gtk.TextTag indent_tag = new Gtk.TextTag (null);
+				indent_tag.Indent = -10;
+				indent_tag.Data ["has_indent"] = true;
+
+				Buffer.TagTable.Add (indent_tag);
+				Buffer.ApplyTag (indent_tag, insert, para_end);
+				}
+				break;
+				*/
+				
 			case Gdk.Key.Tab:
 			case Gdk.Key.KP_Tab:
 				Console.WriteLine ("Got Tab!");
@@ -902,7 +1247,7 @@ namespace Tomboy
 				bool indent_exists = false;
 
 				foreach (Gtk.TextTag tag in insert.Tags) {
-					if (tag.Data ["has_indent"] != null) {
+					if (tag.Data ["has_margin"] != null) {
 						Console.WriteLine ("Has Indent Already! Updating!");
 						tag.LeftMargin = tag.LeftMargin + 40;
 						indent_exists = true;
@@ -913,7 +1258,7 @@ namespace Tomboy
 				if (!indent_exists) {
 					Gtk.TextTag indent_tag = new Gtk.TextTag (null);
 					indent_tag.LeftMargin = 40;
-					indent_tag.Data ["has_indent"] = true;
+					indent_tag.Data ["has_margin"] = true;
 
 					Buffer.TagTable.Add (indent_tag);
 					Buffer.ApplyTag (indent_tag, insert, para_end);
@@ -935,7 +1280,7 @@ namespace Tomboy
 				para_end2.ForwardToLineEnd ();
 
 				foreach (Gtk.TextTag tag in insert2.Tags) {
-					if (tag.Data ["has_indent"] != null) {
+					if (tag.Data ["has_margin"] != null) {
 						Console.WriteLine ("Has Indent Already! Updating!");
 						tag.LeftMargin = tag.LeftMargin - 40;
 						if (tag.LeftMargin == 0) {
@@ -950,6 +1295,60 @@ namespace Tomboy
 				break;
 			}
 		}
+
+		void OnInsertText (object sender, Gtk.InsertTextArgs args)
+		{
+			Console.WriteLine ("Trying hanging indent! line-offset:{0}", args.Pos.LineOffset);
+
+			if (args.Text != "*" || args.Pos.LineOffset != 1)
+				return;
+
+			Gtk.TextIter para_start = args.Pos;
+			para_start.BackwardChar ();
+
+			Gtk.TextTag indent_tag = new Gtk.TextTag (null);
+			indent_tag.Indent = -10;
+			indent_tag.Data ["has_indent"] = true;
+
+			Buffer.TagTable.Add (indent_tag);
+			Buffer.ApplyTag (indent_tag, para_start, args.Pos);
+		}
+
+		// 
+		// <list>
+		// <anonymous list-prefix="124">Content content content</anonymous>
+		// <anonymous indent="40">
+		//   <anonymous list-prefix="123">
+		//     Content content content
+		//   </anonymous>
+		// </anonymous>
+		// </list>
+
+		string GetListPrefix (Gtk.TextIter iter)
+		{
+			Gtk.TextTag number_list = Buffer.TagTable.Lookup ("number-list");
+			Gtk.TextTag bullet_list = Buffer.TagTable.Lookup ("bullet-list");
+
+			if (iter.HasTag (bullet_list))
+				return "* ";
+
+			int add = 0;
+
+			while (iter.HasTag (number_list)) {
+				foreach (Gtk.TextTag tag in iter.Tags) {
+					if (tag.Data ["list-prefix"] != null) {
+						int last_num = (int) tag.Data ["list-prefix"];
+						last_num += add;
+						return last_num.ToString () + ". ";
+					}
+				}
+
+				iter.BackwardLine ();
+				add++;
+			}
+
+			return null;
+		}
 	}
 
 #if BROKEN
